import numpy as np
import Training_data

rng = np.random.default_rng(123)
TEACHDATA = 1000
TESTDATA = 100
# ETA = 0.5
T_NUMBER = 3  # Number to be detected 0-6


def sigmoid(val):
    return 1 / (1 + np.exp(-val))


def linear_act(val):

    # there is the option to make it linear in the valid range and make it constant otherwise
    return val


def threshold(val):
    return 1 if val > 0 else 0


class Neuron:
    test_data = Training_data.make_testset(TESTDATA)
    teach_data = Training_data.make_testset(TEACHDATA)

    def __init__(self, input_count, activation):
        self.input_count = input_count
        self.activation = activation
        self.weights = rng.random(input_count + 1)

    def test(self):
        res = [0 for _ in range(len(Neuron.test_data))]
        for number in range(len(Neuron.test_data)):
            for sample in Neuron.test_data[number]:
                ix = np.insert(sample.ravel(), 0, 1)
                res[number] = res[number] + abs(self.activation(ix.dot(self.weights)))
        return res


class ThresholdPerceptron(Neuron):
    def __init__(self, input_count, activation=threshold):
        super().__init__(input_count, activation)

    def train(self, ETA):
        for i in range(TEACHDATA):
            old_weights = np.copy(self.weights)
            for j in rng.permutation(len(Neuron.teach_data)):
                T = 1 if j == T_NUMBER else 0
                ix = np.insert(Neuron.teach_data[j][i].ravel(), 0, 1)
                RI = self.activation(ix.dot(self.weights))
                if RI != T:
                    delta = ETA * \
                            (T - self.activation(ix.dot(self.weights))) * ix
                    self.weights = self.weights + delta
            if np.linalg.norm(old_weights - self.weights) == 0.00:
                return self.weights
        return self.weights


class SGDPerceptron(Neuron):
    def __init__(self, input_count, activation=sigmoid):
        super().__init__(input_count, activation)

    def train(self, ETA):

        for i in range(TEACHDATA):
            old_weights = np.copy(self.weights)
            delta = [0 for _ in range(len(old_weights))]
            for j in rng.choice(rng.permutation(len(Neuron.teach_data)),3):
                T = (j == T_NUMBER)
                ix = np.insert(Neuron.teach_data[j][i].ravel(), 0, 1)
                z = ix.dot(self.weights)
                RI = self.activation(z)
                delta = ETA * (T - RI) * RI * (1 - RI) * ix
            self.weights += delta
        return self.weights


class LinearPerceptron(Neuron):
    def __init__(self, input_count, activation=linear_act):
        super().__init__(input_count, activation)

    def train(self, ETA):
        for i in range(TEACHDATA):
            old_weights = np.copy(self.weights)
            delta = [0 for _ in range(len(old_weights))]
            for j in rng.permutation(len(Neuron.teach_data)):
                T = (j == T_NUMBER)
                ix = np.insert(Neuron.teach_data[j][i].ravel(), 0, 1)
                delta += ETA * (T - self.activation(ix.dot(self.weights))) * ix
            self.weights = self.weights + delta
            if np.linalg.norm(old_weights - self.weights) == 0.00:
                return self.weights
        return self.weights



